from windML.oneshot.models.extreme_learning_machine import ELMClassifier
from windML.oneshot.modelfree.matrix_factorisation import MFRClassifier
from windML.fewshot.direct_feedback_alignment.dfa_classifier import DFAClassifier
from windML.fewshot.recurrent_neural_network.character_rnn import charRNN
from windML.fewshot.recurrent_neural_network.rnn import RNN

__version__ = "0.2.3"
